package anywheresoftware.b4a.agraham.reflection;

import java.lang.ref.WeakReference;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.HashMap;

import android.app.Activity;
import android.content.Context;
import android.util.Log;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.View;
import android.view.ContextMenu.ContextMenuInfo;
import android.view.View.OnClickListener;
import android.view.View.OnCreateContextMenuListener;
import android.view.View.OnFocusChangeListener;
import android.view.View.OnKeyListener;
import android.view.View.OnLongClickListener;
import android.view.View.OnTouchListener;
import android.view.ContextMenu;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.BA.Author;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;

/**
 *This is the object that does the reflection. In order to use this successfully you will
 *need an understanding of the use of Java classes and their fields and methods.
 * 
 *'Full' (it is often lacking useful explanatory details) technical documentation is available
 *on the Google Android website at http://developer.android.com.
 *
 *Java is case sensitive and as used for Android does not support properties. Properties as
 *implemented in Basic4Android are actually methods with lower case prefixes 'set' and 'get'.
 *'set' methods take a single parameter and return void, 'get' methods take no parameters and
 *return the requested values. Any other method signatures are exposed by Basic4Android as
 *normal methods. For example the Top property of a View is actually implemented in Java code as
 *two methods, int getLeft() and void setLeft(int left). The Basic4Android compiler makes them
 *look like a single property to the programmer.
 */
@ShortName("Reflector")
@Author("Andrew Graham")
@Version(2.2f)
public class Reflection
{
	// 1.0 initial release
	// 1.1 added RunStaticMethod
	// 1.2 CreateObject2 was passing types not args to the constructor
	// 1.3 Accept nulls for args and types in RunMethod and RunStaticMethod
	//	   RunStaticMethod and runmethod were confusingly returning the last Exception caught not the first
	// 1.4 Added library comment about primitive types specifications
	//	   1.3 mod to RunStaticMethod and runmethod didn't throw NoSuchMethodException
	//	   Added GetMethod and InvokeMethod
	// 1.5 added GetContext
	// 1.6 added TargetRank, GetArray, SetArray, SetArray2, IsNull, SetField2
	//	   added GetB4AClass, GetProcessBA, GetMostCurrent, GetActivity, GetActivityBA
	//	   Reflection is no longer an ActivitytObject
	// 1.7 added GetStaticField, SetStaticField and SetStaticField2
	//	   added GetPublicField, SetPublicField and SetPublicField2
	//	   added GetFieldInfo, GetField2, SetField3 and SetField4
	//	   added RunPublicMethod
	// 1.8 added GetProxy
	// 1.9 added shorthand for primitive type specification
	//	   added explanation of array syntax to LIBRARY_DOC
	//	   changed exception handling of GetB4AClass
	//	   corrected SetStaticField and SetStaticField2 to void return
	//	   added KeyEvent and MotionEvent parameter to OnKey and OnTouch event Sub parameters
	// 2.0 added check for null return from RaiseEvent in onLongClick and onKey listeners
	// 2.1 added warning about DoEvents in SetOnTouchListener and SetOnFocusListener
	//	   runmethod now loooks at Object class to allow getClass in particular
	// 2.2 GetActivity and GetActivityBA changed for Basic4android 2.00 and higher
	private final static double version = 2.2;

	/**
	 *This library contains a Reflector object that allows access to methods and fields
	 *of Android objects that are not exposed to the Basic4Android language. It does this
	 *by means of a facility called "Reflection" that uses meta-data for objects that is
	 *included in the application package and allows dynamic access to fields and methods
	 *at runtime.
	 *
	 *GetActivity and GetActivityBA in this library only work for Basic4android 2.00 and higher.
	 *Use version 2.1 of this library for earlier versions of Basic4android.
	 *
	 *The specifications for the primitive Java types recognised by this object are
	 *"java.lang.boolean", "java.lang.byte", "java.lang.char", "java.lang.short",
	 *"java.lang.int", "java.lang.long", "java.lang.float", and "java.lang.double"
	 *Note the lower-cased names. The single uppercase letter representation, as
	 *listed for arrays below, is also recognised.
	 *
	 *The specifications for boxed primitive Java types which are true objects are
	 *"java.lang.Boolean", "java.lang.Byte", "java.lang.Char", "java.lang.Short",
	 *"java.lang.Int", "java.lang.Long", "java.lang.Float", and "java.lang.Double"
	 *Note the upper-cased names.
	 *
	 *The specifications for arrays of primitive types is a number of square left brackets
	 *corresponding to the rank of the array followed by a single letter indicating the type.
	 *  B represents the byte primitive type, S represents the short primitive type,
	 *  I represents the int primitive type, J represents the long primitive type,
	 *  F represents the float primitive type, D represents the double primitive type,
	 *  C represents the char primitive type, Z represents the boolean primitive type.
	 *So a two dimension integer array is represented by "[[I".
	 *
	 *The specifications for arrays of objects is similar but with the single letter replaced
	 *the letter "L", followed by the class name terminated by a semicolon ";".
	 *So a single dimension array of strings is represented by "[Ljava.lang.String;"
	 */
	public void LIBRARY_DOC()
	{
	}

	/**
	 *This field holds the object that is being reflected upon. The target object is
	 *assigned to this field where it can then be manipulated as required. 
	 */
	public Object Target;

	private static HashMap<String, Class<?>> primitives = new HashMap<String, Class<?>>()
	{
		private static final long serialVersionUID = 1L;
		{
			put("java.lang.boolean", Boolean.TYPE); // 1 bit size undefined boolean
			put("java.lang.byte", Byte.TYPE); // 8 bit signed integer
			put("java.lang.char", Character.TYPE); // 16 bit Unicode char
			put("java.lang.short", Short.TYPE); // 16 bit signed integer
			put("java.lang.int", Integer.TYPE); // 32 bit signed integer
			put("java.lang.long", Long.TYPE); // 64 bit signed integer
			put("java.lang.float", Float.TYPE); // 32 bit signed number
			put("java.lang.double", Double.TYPE); // 64 bit signed number
			put("Z", Boolean.TYPE); // 1 bit size undefined boolean
			put("B", Byte.TYPE); // 8 bit signed integer
			put("C", Character.TYPE); // 16 bit Unicode char
			put("S", Short.TYPE); // 16 bit signed integer
			put("I", Integer.TYPE); // 32 bit signed integer
			put("J", Long.TYPE); // 64 bit signed integer
			put("F", Float.TYPE); // 32 bit signed number
			put("D", Double.TYPE); // 64 bit signed number
		}
	};

	// static methods

	private static Class<?> classforname(String type)
	{
		if (primitives.containsKey(type))
		{
			return primitives.get(type);
		} else
			try
			{
				return Class.forName(type);
			} catch (ClassNotFoundException e)
			{
				throw new RuntimeException(e);
			}
	}

	@SuppressWarnings("unchecked")
	private static Object sethelper(String o, Class t)
	{
		if (t.isEnum())
			return Enum.valueOf(t, (String) o);
		if (t.isPrimitive())
		{
			if (t == Boolean.TYPE)
				return Boolean.parseBoolean((String) o);
			else if (t == Byte.TYPE)
				return Byte.parseByte((String) o);
			else if (t == Character.TYPE)
				return ((String) o).charAt(0);
			else if (t == Short.TYPE)
				return Short.parseShort((String) o);
			else if (t == Integer.TYPE)
				return Integer.parseInt((String) o);
			else if (t == Long.TYPE)
				return Long.parseLong((String) o);
			else if (t == Float.TYPE)
				return Float.parseFloat((String) o);
			else if (t == Double.TYPE)
				return Double.parseDouble((String) o);
		}
		return o;
	}

	private Object runmethod(String method, Object[] Args, Class<?>[] types) throws Exception
	{
		Exception ex = null;
		Class<?> ob = Class.forName("java.lang.Object");
		Class<?> ot = Target.getClass();
		Boolean loop = true;
		do
		{
			Method m;
			try
			{
				if (ot == ob)
					loop = false;
				m = ot.getDeclaredMethod(method, types);
				m.setAccessible(true); // only needed for private methods
				return m.invoke(Target, Args);
			} catch (NoSuchMethodException e)
			{
				ex = e;
				ot = ot.getSuperclass();
			} catch (InvocationTargetException e)
			{
				throw (Exception) e.getCause();
			}
		} while (loop);
		throw ex;
	}

	// Public methods

	/**
	 *Creates and returns a new object of the specified type using the default constructor.
	 */
	public Object CreateObject(String type)
	{
		try
		{
			Class<?> ot = classforname(type);
			return ot.newInstance();
		} catch (Exception e)
		{
			throw new RuntimeException(e);
		}
	}

	/**
	 *Creates and returns new object of the specified type using the constructor that matches
	 *the array of type names given and passing it the arguments provided.
	 *
	 *The array of type names is needed in order to find the correct constructor because
	 *primitives passed in the Args array are boxed and so CreateNew cannot tell whether
	 *to look for a target constructor that accepts a primitive parameter type or a boxed
	 *primitive object type.
	 */
	public Object CreateObject2(String type, Object[] args, String[] types)
	{
		Class<?>[] ts = new Class[args.length];
		Constructor<?> ci;
		try
		{
			Class<?> ot = classforname(type);
			for (int i = 0; i < args.length; i++)
				ts[i] = classforname(types[i]);
			ci = ot.getConstructor(ts);
			return ci.newInstance((Object[]) args);
		} catch (Exception e)
		{
			throw new RuntimeException(e);
		}
	}

	// FIELD METHODS

	/**
	 *Finding a field from its string representation is expensive so this method can be used
	 *to get the Field information object and save it for multiple accesses of the same field.
	 *Protected and private fields may be accessed if allowed by any security manager which may be present.
	 */
	public Field GetFieldInfo(String field) throws Exception
	{
		Exception ex;
		Class<?> ob = Class.forName("java.lang.Object");
		Class<?> ot = Target.getClass();
		do
		{
			Field f;
			try
			{
				f = ot.getDeclaredField(field);
				f.setAccessible(true); // only needed for private fields
				return f;
			} catch (Exception e)
			{
				ex = e;
				ot = ot.getSuperclass();
			}
		} while (ot != ob);
		throw ex;
	}

	/**
	 *Returns the value of the field of the current target. Protected and private fields
	 *may be accessed if allowed by any security manager which may be present.
	 *Target must be an instance of a Class, not a Class object.
	 */
	public Object GetField(String field) throws Exception
	{
		Exception ex;
		Class<?> ob = Class.forName("java.lang.Object");
		Class<?> ot = Target.getClass();
		do
		{
			Field f;
			try
			{
				f = ot.getDeclaredField(field);
				f.setAccessible(true); // only needed for private fields
				return f.get(Target);
			} catch (Exception e)
			{
				ex = e;
				ot = ot.getSuperclass();
			}
		} while (ot != ob);
		throw ex;
	}

	/**
	 *Returns the value of the field of the current target.
	 *Target must be an instance of a Class, not a Class object.
	 */
	public Object GetField2(Field fieldinfo) throws Exception
	{
		return fieldinfo.get(Target);
	}

	/**
	 *Returns the value of the public field of the current target.
	 *This is more efficient than GetField but can only access public fields.
	 *Target must be an instance of a Class, not a Class object.
	 */
	public Object GetPublicField(String field)
	{
		try
		{
			Class<?> ot = Target.getClass();
			Field f = ot.getField(field);
			return f.get(Target);
		} catch (Exception e)
		{
			throw new RuntimeException(e);
		}
	}

	/**
	 *Returns the value of the specified static field of the specified class.
	 *Protected and private fields may be accessed if allowed by any security manager
	 *which may be present.
	 *Static fields may also be accessed with GetField and an instance of the class.
	 */
	public Object GetStaticField(String classname, String field) throws Exception
	{
		Exception ex = null;
		Class<?> ob = Class.forName("java.lang.Object");
		Class<?> ot = Class.forName(classname);
		do
		{
			Field f;
			try
			{
				f = ot.getDeclaredField(field);
				f.setAccessible(true); // only needed for private fields
				return f.get(null);
			} catch (Exception e)
			{
				ex = e;
				ot = ot.getSuperclass();
			}
		} while (ot != ob);
		throw ex;
	}

	/**
	 *Sets the specified field of the current target to the value provided. Protected and
	 *private fields may be accessed if allowed by any security manager which may be present.
	 *Target must be an instance of a Class, not a Class object.
	 */
	public void SetField(String field, String value, String type) throws Exception
	{
		Exception ex;
		Class<?> ob = Class.forName("java.lang.Object");
		Class<?> ot = Target.getClass();
		do
		{
			Field f;
			try
			{
				f = ot.getDeclaredField(field);
				f.setAccessible(true); // only needed for private fields
				Object o = sethelper(value, classforname(type));
				f.set(Target, o);
				return;
			} catch (Exception e)
			{
				ex = e;
				ot = ot.getSuperclass();
			}
		} while (ot != ob);
		throw ex;
	}
	
	/**
	 *Sets the specified field of the current target to the value provided. Protected and
	 *private fields may be accessed if allowed by any security manager which may be present.
	 *Target must be an instance of a Class, not a Class object.
	 */
	public void SetField2(String field, Object value) throws Exception
	{
		Exception ex;
		Class<?> ob = Class.forName("java.lang.Object");
		Class<?> ot = Target.getClass();
		do
		{
			Field f;
			try
			{
				f = ot.getDeclaredField(field);
				f.setAccessible(true); // only needed for private fields
				f.set(Target, value);
				return;
			} catch (Exception e)
			{
				ex = e;
				ot = ot.getSuperclass();
			}
		} while (ot != ob);
		throw ex;
	}
	
	/**
	 *Sets the specified field of the current target to the value provided.
	 *Target must be an instance of a Class, not a Class object.
	 */
	public void SetField3(Field fieldinfo, String value, String type) throws Exception
	{
		Object o = sethelper(value, classforname(type));
		fieldinfo.set(Target, o);
	}

	/**
	 *Sets the specified field of the current target to the value provided.
	 *Target must be an instance of a Class, not a Class object.
	 */
	public void SetField4(Field fieldinfo, Object value) throws Exception
	{
		fieldinfo.set(Target, value);
	}

	

	/**
	 *Sets the specified field of the current target to the value provided.
	 *This is more efficient than SetField but can only access public fields.
	 *Target must be an instance of a Class, not a Class object.
	 */
	public void SetPublicField(String field, String value, String type)
	{
		try
		{
			Class<?> ot = Target.getClass();
			Field f = ot.getField(field);
			Object o = sethelper(value, classforname(type));
			f.set(Target, o);
		} catch (Exception e)
		{
			throw new RuntimeException(e);
		}
	}
	

	/**
	 *Sets the specified field of the current target to the value provided.
	 *This is more efficient than SetField but can only access public fields.
	 *Target must be an instance of a Class, not a Class object.
	 */
	public void SetPublicField2(String field, Object value)
	{
		try
		{
			Class<?> ot = Target.getClass();
			Field f = ot.getField(field);
			f.set(Target, value);
		} catch (Exception e)
		{
			throw new RuntimeException(e);
		}
	}
	

	/**
	 *Sets the specified static field of the specified class to the value provided.
	 *Protected and private fields may be accessed if allowed by any security manager
	 *which may be present.
	 *Static fields may also be accessed with SetField and an instance of the class.
	 */
	public void SetStaticField(String classname, String field, String value, String type) throws Exception
	{
		Exception ex = null;
		Class<?> ob = Class.forName("java.lang.Object");
		Class<?> ot = Class.forName(classname);
		do
		{
			Field f;
			try
			{
				f = ot.getDeclaredField(field);
				f.setAccessible(true); // only needed for private fields
				Object o = sethelper(value, classforname(type));
				f.set(null, o);
				return;
			} catch (Exception e)
			{
				ex = e;
				ot = ot.getSuperclass();
			}
		} while (ot != ob);
		throw ex;
	}
	

	/**
	 *Sets the specified static field of the specified class to the value provided.
	 *Protected and private fields may be accessed if allowed by any security manager
	 *which may be present.
	 *Static fields may also be accessed with SetField and an instance of the class.
	 */
	public void SetStaticField2(String classname, String field, Object value) throws Exception
	{
		Exception ex = null;
		Class<?> ob = Class.forName("java.lang.Object");
		Class<?> ot = Class.forName(classname);
		do
		{
			Field f;
			try
			{
				f = ot.getDeclaredField(field);
				f.setAccessible(true); // only needed for private fields
				f.set(null, value);
				return;
			} catch (Exception e)
			{
				ex = e;
				ot = ot.getSuperclass();
			}
		} while (ot != ob);
		throw ex;
	}
	

	// ARRAY METHODS

	/**
	 *Returns an int array whose length is the number of dimensions of the array and whose contents
	 *are the length of the first element of each array dimension.
	 *A zero length integer array is returned if Target is not an array.
	 */
	public int[] TargetRank()
	{
		Class<?> type = Target.getClass();
		if (!type.isArray())
			return new int[0];
		else
		{
			String clsName = type.getName();
			int dims = 1 + clsName.lastIndexOf('[');
			int[] lens = new int[dims];
			Object arr = Target;
			for (int i = 0; i < dims; i++)
			{
				int len = Array.getLength(arr);
				lens[i] = len;
				if (0 < len)
				{
					arr = Array.get(arr, 0);
				}
			}
			return lens;
		}
	}
	

	/**
	 *Returns the Object at the position(s) in an array specified by the contents of indeces.
	 *indeces must be an integer array of the same rank as the Target array or an error will occur.
	 */
	public Object GetArray(int[] indeces)
	{
		Object arr = Target;
		for (int i = 0; i < indeces.length - 1; i++)
		{
			arr = Array.get(arr, indeces[i]);
		}
		return Array.get(arr, indeces[indeces.length - 1]);
	}
	

	/**
	 *Set the position(s) in an array specified by the contents of indeces to the spcified value.
	 *indeces must be an integer array of the same rank as the Target array or an error will occur.
	 */
	public void SetArray(int[] indeces, String value, String type)
	{
		Object arr = Target;
		for (int i = 0; i < indeces.length - 1; i++)
		{
			arr = Array.get(arr, indeces[i]);
		}
		Object o = sethelper(value, classforname(type));
		Array.set(arr, indeces[indeces.length - 1], o);
	}
	

	/**
	 *Set the position(s) in an array specified by the contents of indeces to the specified value.
	 *indeces must be an integer array of the same rank as the Target array or an error will occur.
	 */
	public void SetArray2(int[] indeces, Object value)
	{
		Object arr = Target;
		for (int i = 0; i < indeces.length - 1; i++)
		{
			arr = Array.get(arr, indeces[i]);
		}
		Array.set(arr, indeces[indeces.length - 1], value);
	}
	

	// CONTEXT AND DEBUG METHODS

	/**
	 *Returns the Context of the Process to which the Reflection object belongs.
	 *This is the Application object returned from Activity.getApplicationContext().
	 */
	public Context GetContext(BA ba)
	{
		// ba is always a process BA
		return ba.context;
	}
	

	/**
	 *Returns the current activity if any.
	 *To avoid memory leaks this should not be used by a Reflector that is a Process object.
	 *To use this requires a knowledge of the structure of a Basic4android application.
	 */
	public Activity GetActivity(BA ba)
	{
		// ba is always a process BA
		WeakReference<BA> w = ba.sharedProcessBA.activityBA;
		return w.get().activity;
	}
	

	/**
	 *Returns the Activity BA of the current activity.
	 *To avoid memory leaks this should not be used by a Reflector that is a Process object.
	 *To use this requires a knowledge of the structure of a Basic4android application.
	 */
	public BA GetActivityBA(BA ba)
	{
		// ba is always a process BA
		WeakReference<BA> w = ba.sharedProcessBA.activityBA;
		return w.get();
	}
	

	/**
	 *Returns the Java Class for the specified B4A Activity, Service or Code module.
	 *To use this requires a knowledge of the structure of a Basic4android application.
	 */
	public Class<?> GetB4AClass(String component)
	{
		Class<?> c;
		try
		{
			c = Class.forName(BA.packageName + "." + ((String) component).toLowerCase(BA.cul));
		} catch (Exception e)
		{
			throw new RuntimeException ("Class " + component + " not found.");
			// Cause of ClassNotFoundException is a NoClassDefFoundError 
			// BA.setLastException can't handle an Error as a Cause
			//throw new RuntimeException(e);
		}
		return c;
	}
	

	/**
	 *Returns the processBA instance for the specified B4A Activity or Service module.
	 *To use this requires a knowledge of the structure of a Basic4android application.
	 */
	public BA GetProcessBA(String component)
	{
		Class<?> c;
		Field f;
		try
		{
			c = Class.forName(BA.packageName + "." + ((String) component).toLowerCase(BA.cul));
			f = c.getDeclaredField("processBA");
			f.setAccessible(true);
			return (BA) f.get(null);
		} catch (Exception e)
		{
			throw new RuntimeException(e);
		}
	}
	

	/**
	 *Returns the current instance for the specified B4A Activity or Service module.
	 *This might return null if the Activity or Service is not instantiated.
	 *Note that Code modules do not have a current instance.
	 *To use this requires a knowledge of the structure of a Basic4android application.
	 */
	public Object GetMostCurrent(String component)
	{
		Class<?> c;
		Field f;
		try
		{
			c = Class.forName(BA.packageName + "." + component.toLowerCase(BA.cul));
			f = c.getDeclaredField("mostCurrent");
			f.setAccessible(true);
			return f.get(null);
		} catch (Exception e)
		{
			throw new RuntimeException(e);
		}
	}
	

	// METHOD METHODS

	/**
	 *Finding a method from its string representation is expensive so this method can be used
	 *to get the Method information object and save it for multiple invocations of the same method.
	 * 
	 *The String array of type names is needed in order to find the correct variant of the method.
	 */
	public Method GetMethod(String method, String[] types) throws Exception
	{
		Exception ex = null;
		Class<?> ob = Class.forName("java.lang.Object");
		Class<?> ot = Target.getClass();
		if (types == null)
			types = new String[0];
		Class<?>[] ots = new Class[types.length];
		for (int i = 0; i < types.length; i++)
		{
			ots[i] = classforname(types[i]);
		}
		do
		{
			Method m;
			try
			{
				m = ot.getDeclaredMethod(method, ots);
				m.setAccessible(true); // only needed for private methods
				return m;
			} catch (NoSuchMethodException e)
			{
				ex = e;
				ot = ot.getSuperclass();
			}
		} while (ot != ob);
		throw ex;
	}

	/**
	 *Invoke the provided Method on the provided object instance and return the result.
	 */
	public Object InvokeMethod(Object instance, Method method, Object[] args) throws Exception
	{
		try
		{
			return method.invoke(instance, args);
		} catch (InvocationTargetException e)
		{
			throw (Exception) e.getCause();
		}

	}
	

	/**
	 *Runs the specified method on the current target. Protected and private methods
	 *may be accessed if allowed by any security manager which may be present.
	 */
	public Object RunMethod(String method) throws Exception
	{
		return runmethod(method, new Object[0], new Class[0]);
	}
	

	/**
	 *Runs the specified method on the current target passing it the argument provided.
	 *Protected and private methods may be accessed if allowed by any security manager
	 *which may be present.
	 *	 */
	public Object RunMethod2(String method, String arg1, String type1) throws Exception
	{
		Object[] oo = new Object[1];
		Class<?>[] ot = new Class[1];
		ot[0] = classforname(type1);
		oo[0] = sethelper(arg1, ot[0]);
		return runmethod(method, oo, ot);
	}
	

	/**
	 *Runs the specified method on the current object passing it the arguments provided.
	 *Protected and private methods may be accessed if allowed by any security manager
	 *which may be present.
	 */
	public Object RunMethod3(String method, String arg1, String type1, String arg2, String type2) throws Exception
	{
		Object[] oo = new Object[2];
		Class<?>[] ot = new Class[2];
		ot[0] = classforname(type1);
		ot[1] = classforname(type2);
		oo[0] = sethelper(arg1, ot[0]);
		oo[1] = sethelper(arg2, ot[1]);
		return runmethod(method, oo, ot);
	}
	

	/**
	 *Runs the specified method on the current target passing it the arguments provided.
	 *Protected and private methods may be accessed if allowed by any security manager
	 *which may be present.
	 *
	 *The String array of type names is needed in order to find the correct method because
	 *primitives passed in the Args array are boxed and so RunMethod cannot tell whether to
	 *look for a target method that accepts a primitive parameter type or a boxed primitive
	 *object type.
	 */
	public Object RunMethod4(String method, Object[] args, String[] types) throws Exception
	{
		if (types == null)
			types = new String[0];
		Class<?>[] ot = new Class[types.length];
		for (int i = 0; i < types.length; i++)
		{
			ot[i] = classforname(types[i]);
		}
		return runmethod(method, args, ot);
	}
	

	/**
	 *Runs the specified method on the current target passing it the arguments provided.
	 *This is more efficient that RunMethd4 but the method must be public.
	 *
	 *The String array of type names is needed in order to find the correct method because
	 *primitives passed in the Args array are boxed and so RunMethod cannot tell whether to
	 *look for a target method that accepts a primitive parameter type or a boxed primitive
	 *object type.
	 */
	public Object RunPublicmethod(String Method, Object[] Args, Class<?>[] types)
	{
		Class<?> ot = Target.getClass();
		try
		{
			Method m = ot.getMethod(Method, types);
			return m.invoke(Target, Args);
		} catch (Exception e)
		{
			throw new RuntimeException(e);
		}
	}
	

	/**
	 *Runs the specified static method of the specified class passing it the arguments provided.
	 *Protected and private methods may be accessed if allowed by any security manager
	 *which may be present.
	 *
	 *The String array of type names is needed in order to find the correct method because
	 *primitives passed in the Args array are boxed and so RunMethod cannot tell whether to
	 *look for a target method that accepts a primitive parameter type or a boxed primitive
	 *object type. For methods that take no parameters Null may passed for args and types.
	 */
	public Object RunStaticMethod(String classname, String method, Object[] args, String[] types) throws Exception
	{
		Exception ex = null;
		if (types == null)
			types = new String[0];
		Class<?> ob = Class.forName("java.lang.Object");
		Class<?> ot = Class.forName(classname);
		Class<?>[] ots = new Class[types.length];
		for (int i = 0; i < types.length; i++)
		{
			ots[i] = classforname(types[i]);
		}
		do
		{
			Method m;
			try
			{
				m = ot.getDeclaredMethod(method, ots);
				m.setAccessible(true); // only needed for private methods
				return m.invoke(null, args);
			} catch (NoSuchMethodException e)
			{
				ex = e;
				ot = ot.getSuperclass();
			} catch (InvocationTargetException e)
			{
				throw (Exception) e.getCause();
			}
		} while (ot != ob);
		throw ex;
	}
	

	/**
	 *Returns the result of running the "toString()" method of the current object.
	 */
	public String ToString()
	{
		return Target.toString();
	};

	// Public properties

	/**
	 *Returns the version number of the library.
	 */
	public double getVersion()
	{
		return version;
	}
	

	/**
	 *Returns the name of the class of the current object.
	 */
	public String getTypeName()
	{
		return (String) Target.getClass().getName();
	}

	/**
	 *Returns True if the prsent value of Target is Null.
	 */
	public boolean getIsNull()
	{
		return Target == null;
	}
	

	// EVENT METHODS

	/**
	 *Target must be a View of some sort.
	 *In most cases Basic4Android will have already exposed this as a Click event.
	 *Sets the OnClickListener of the view to a Sub that must have a signature of
	 *Sub Whatever(viewtag As Object).
	 */
	public void SetOnClickListener(final BA ba, String sub)
	{
		final String fsub = sub.toLowerCase();
		((View) Target).setOnClickListener(new OnClickListener()
		{
			@Override
			public void onClick(View v)
			{
				ba.raiseEvent(v, fsub, new Object[] { v.getTag() });
			}
		});
	}
	

	/**
	 *Target must be a View of some sort.
	 *In most cases Basic4Android will have already exposed this as a LongClick event.
	 *Sets the OnLongClickListener of the view to a Sub that must have a signature of
	 *Sub Whatever(viewtag As Object) As Boolean.
	 *
	 *This Sub must return True if it wants to consume the event or False otherwise.
	 */
	public void SetOnLongClickListener(final BA ba, String sub)
	{
		final String fsub = sub.toLowerCase();
		((View) Target).setOnLongClickListener(new OnLongClickListener()
		{
			@Override
			public boolean onLongClick(View v)
			{
				//return (Boolean) ba.raiseEvent(v, fsub, new Object[] { v.getTag() });
				Object b = ba.raiseEvent(v, fsub, new Object[] { v.getTag()});
				if (b != null)
				{
					return (Boolean) b;
				}
				else
				{
					Log.w("B4A", fsub  + " raiseEvent returned null");
					return false;					
				}
		}
		});
	}
	

	/**
	 *Target must be a View of some sort.
	 *This is included for completeness of all the Listeners that class View supports.
	 *Sets the OnCreateContextMenuListener of the view to a Sub that must have a signature of
	 *Sub Whatever(viewtag As Object, menu As Object, menuinfo As Object)
	 */
	public void SetOnCreateContextMenuListener(final BA ba, String sub)
	{
		final String fsub = sub.toLowerCase();
		((View) Target).setOnCreateContextMenuListener(new OnCreateContextMenuListener()
		{
			//@Override
			public void onCreateContextMenu(ContextMenu menu, View v, ContextMenuInfo menuInfo)
			{
				ba.raiseEvent(v, fsub, new Object[] { v.getTag(), menu, menuInfo });
			}
		});
	}
	

	/**
	 *Target must be a View of some sort.
	 *Sets the onFocusChangeListener of the view to a Sub that must have a signature of
	 *Sub Whatever(viewtag As Object, focus As Boolean).
	 *
	 *You should make sure not to call DoEvents, Msgbox or any modal Dialog inside this
	 *event as it will fail in Android 4.0.3 and above. 
	 *
	 *It may also fail if Debug is paused in the event in Android 4.0.3 and above. 
 	 */
	public void SetOnFocusListener(final BA ba, String sub)
	{
		final String fsub = sub.toLowerCase();
		((View) Target).setOnFocusChangeListener(new OnFocusChangeListener()
		{
			@Override
			public void onFocusChange(View v, boolean hasFocus)
			{
				ba.raiseEvent(v, fsub, new Object[] { v.getTag(), hasFocus });
			}
		});
	}
	

	/**
	 *Target must be a View of some sort.
	 *Sets the onKeyListener of the view to a Sub that must have a signature of
	 *Sub Whatever(viewtag As Object, keycode As Int, keyevent As Object) As Boolean.
	 *
	 *This Sub must return True if it wants to consume the event or False otherwise.
	 */
	public void SetOnKeyListener(final BA ba, String sub)
	{
		final String fsub = sub.toLowerCase();
		((View) Target).setOnKeyListener(new OnKeyListener()
		{
			@Override
			public boolean onKey(View v, int keyCode, KeyEvent event)
			{
				Object b = ba.raiseEvent(v, fsub, new Object[] { v.getTag(), keyCode, event });
				if (b != null)
				{
					return (Boolean) b;
				}
				else
				{
					Log.w("B4A", fsub  + " raiseEvent returned null");
					return false;					
				}
			}
		});
	}
	

	/**
	 *Target must be a View of some sort.
	 *Sets the onTouchListener of the view to a Sub that must have a signature of
	 *Sub Whatever (viewtag As Object, action As Int, X As Float, Y As Float, motionevent As Object) As Boolean.
	 *
	 *This Sub must return True if it wants to consume the event or False otherwise.
	 *
	 *You should make sure not to call DoEvents, Msgbox or any modal Dialog inside this
	 *event as it will fail in Android 4.0.3 and above. 
	 *
	 *It may also fail if Debug is paused in the event in Android 4.0.3 and above. 
	 */
	public void SetOnTouchListener(final BA ba, String sub)
	{
		final String fsub = sub.toLowerCase();
		((View) Target).setOnTouchListener(new OnTouchListener()
		{
			@Override
			public boolean onTouch(View v, MotionEvent event)
			{
				return (Boolean) ba.raiseEvent(v, fsub, new Object[] { v.getTag(), event.getAction(), event.getX(), event.getY(), event });
			}
		});
	}
	

	
	// Proxy handling
	
	
	/**
	 *In Java you can generate an interface at runtime and have it run a pre-compiled method.
	 *Many events in Android are handled by an interface that typically has an "onXxxxx" method
	 *that is called with some parameters relevant to the event. The interface is typically
	 *set on an object using that objects "setOnXxxxxListener" method.
	 *
	 *This GetProxy method dynamically creates a proxy instance that implements one or more
	 *specified interfaces and which contains the code to call a specified Basic4android Sub
	 *when any of the interface methods are called.
	 *
	 *Typically this instance will implement one or more listeners and will then be assigned
	 *to an object instance using RunMethod4 and its setOnXxxxxListener method.
	 *
	 *When a method of one of the specified interfaces is called the proxy will call the specified
	 *Basic4android Sub passing the method name as a string and any arguments in an object array.
	 *
	 *Note that interfaces declared as internal to a class will need a "$" instead of a "." as their
	 *final separator and all interfaces need to be fully qualified.
	 *e.g android.view.View$OnTouchListener
	 *
	 *The Basic4android Sub called must have the signature
	 *Sub WhateverName(method As String, anyargs() As Object) As Object
	 */
	@SuppressWarnings("unchecked")
	public Proxy GetProxy(BA ba, String[] interfacenames, String b4asubname) throws ClassNotFoundException
	{
		Proxy p;
		Class[] ifs = new Class[interfacenames.length];
		for (int i = 0; i < interfacenames.length; i++)
			ifs[i] = Class.forName(interfacenames[i]);
		p = (Proxy) Proxy.newProxyInstance(ifs[0].getClassLoader(), ifs, new MyInvocationHandler(ba, b4asubname));
		return p;	
	}
	
	
	private class MyInvocationHandler implements InvocationHandler
	{
		String event;
		BA ba;
		public MyInvocationHandler(BA ba, String b4asub)
		{
			event = b4asub.toLowerCase();
			this.ba = ba;
		}
			
		public Object invoke(Object proxy, Method method, Object[] args)
		{
			if (args == null)
				args = new Object[0];
			Object[] args1 = new Object[2];
			args1[0] = method.getName();
			args1[1] = args;
			return ba.raiseEvent(proxy, event, args1);			
		}
	}
	
}
